import sys
from pathlib import Path

import requests
from cargo2rpm.semver import Version

from rust2rpm import log
from rust2rpm.cli import CliArgs, Options
from rust2rpm.conf import TomlConf, load_config_ini, load_config_toml
from rust2rpm.cratesio import NoVersionsError
from rust2rpm.distgit import get_package_info
from rust2rpm.generator import (
    CrateTemplate,
    EpelEightTemplate,
    InvalidMetadataError,
    ProjectTemplate,
    WorkspaceTemplate,
)
from rust2rpm.metadata import WorkspaceError, warn_if_package_uses_restrictive_dependencies
from rust2rpm.patching import PatchError
from rust2rpm.project import CrateProject, InvalidVersionError, LocalProject, OfflineError, Project
from rust2rpm.sysinfo import Target
from rust2rpm.utils import detect_packager, detect_rpmautospec, guess_crate_name


def validate_extra_sources(conf: TomlConf, vendor_tarball: str | None):
    if extra_sources := conf.package.extra_sources:
        min_source_number = 2 if vendor_tarball else 1
        safe_source_number = 2

        for extra_source in extra_sources:
            if extra_source.number == 0:
                log.error(
                    f"Extra source {extra_source.file!r} conflicts with the primary tarball. "
                    f"Use a number larger than {min_source_number} instead.",
                )
                sys.exit(1)
            if extra_source.number == 1 and vendor_tarball:
                log.error(
                    f"Extra source {extra_source.file!r} conflicts with the vendor tarball. "
                    f"Use a number larger than {min_source_number} instead.",
                )
                sys.exit(1)

            # no conflict now, but maybe later
            if extra_source.number < safe_source_number:
                log.warn(
                    f"Extra source {extra_source.file!r} might conflict with other files in the future. "
                    "Using source numbers >= 2 is recommended.",
                )


def validate_extra_patches(conf: TomlConf, *, auto_patch: bool, manual_patch: bool):
    if extra_patches := conf.package.extra_patches:
        min_patch_number = 0
        if auto_patch:
            min_patch_number += 1
        if manual_patch:
            min_patch_number += 1
        safe_patch_number = 2

        for extra_patch in extra_patches:
            if extra_patch.number == 0 and auto_patch:
                log.error(
                    f"Extra patch {extra_patch.file!r} conflicts with the automatically generated patch. "
                    f"Use a number larger than {min_patch_number} instead.",
                )
                sys.exit(1)
            if extra_patch.number == 0 and manual_patch:
                log.error(
                    f"Extra patch {extra_patch.file!r} conflicts with the manually created patch. "
                    f"Use a number larger than {min_patch_number} instead.",
                )
                sys.exit(1)
            if extra_patch.number == 1 and auto_patch and manual_patch:
                log.error(
                    f"Extra patch {extra_patch.file!r} conflicts with the manually created patch. "
                    f"Use a number larger than {min_patch_number} instead.",
                )
                sys.exit(1)

            # no conflict now, but maybe later
            if extra_patch.number < safe_patch_number:
                log.warn(
                    f"Extra patch {extra_patch.file!r} might conflict with other patches in the future. "
                    "Using patch numbers >= 2 is recommended.",
                )


def load_config(
    target: str,
    *,
    path: Path | None = None,
    features: set[str] | None = None,
    migrate: bool = False,
) -> TomlConf:
    ini_paths = [".rust2rpm.conf", "_rust2rpm.conf", "rust2rpm.conf"]
    toml_path = path or Path("rust2rpm.toml")

    # check if both new and old config format are present
    if path is None:
        for ini_path in ini_paths:
            if Path(ini_path).exists() and toml_path.exists():
                log.error(f"Both rust2rpm.toml and {ini_path!s} exist.")
                sys.exit(1)

    tomlconf = load_config_toml(toml_path, features)
    if tomlconf is not None:
        return tomlconf

    if path and tomlconf is None:
        log.error("Configuration file not found: {path!s}")
        sys.exit(1)

    distconf = load_config_ini(ini_paths, target, features)
    if distconf is not None:

        log.warn(
            "The rust2rpm.conf file format is deprecated and support for it will be removed "
            "removed in a future release of rust2rpm. The current settings have been "
            "automatically migrated to the new TOML based format (rust2rpm.toml).",
        )

        # automatically migrate to TOML format unless given explicit path
        if migrate and path is None:
            with toml_path.open("w") as file:
                file.write(distconf.migrate(features))

            for ini_path in ini_paths:
                if Path(ini_path).exists():
                    Path(ini_path).unlink()

        return distconf.upgrade()

    return TomlConf()


def load_project(args: CliArgs, conf: TomlConf) -> Project:
    project: Project

    try:
        if args.path:
            if args.path.is_dir():
                project = LocalProject.from_project_folder(
                    args.path,
                    args.name,
                    args.version,
                    args.output_directory,
                    conf,
                    args.options,
                )
                name = project.name
                return project

            if args.path.name == "Cargo.toml":
                project = LocalProject.from_cargo_toml_path(
                    args.path,
                    args.name,
                    args.version,
                    args.output_directory,
                    conf,
                    args.options,
                )
                name = project.name
                return project

            if args.path.name.endswith(".crate"):
                project = CrateProject.from_local_crate(
                    args.path,
                    args.output_directory,
                    conf,
                    args.options,
                )
                name = project.name
                return project

            msg = (
                "Invalid path argument."
                if args.path.exists()
                else f"Invalid path argument: {args.path!s} is not a local file or directory"
            )
            log.error(msg)
            sys.exit(1)

        else:  # pragma nocover: requires internet access
            if guessed_name := args.name or guess_crate_name():
                name = guessed_name
            else:
                log.error("Neither pkgid nor --path was specified and auto-detection failed.")
                sys.exit(1)

            project = CrateProject.from_crates_io(
                name,
                args.version,
                args.output_directory,
                conf,
                args.options,
            )

    except (InvalidVersionError, NoVersionsError, OfflineError, PatchError, requests.RequestException) as exc:
        log.error(str(exc))
        sys.exit(1)

    else:
        return project


def use_rpmautospec(args: CliArgs, spec_file: Path) -> bool:
    return detect_rpmautospec(args.target, spec_file) if args.options.rpmautospec is None else args.options.rpmautospec


def existence_check(project: Project, args: CliArgs, spec_file: Path):  # pragma nocover: requires internet access
    if (
        args.target in {Target.FEDORA}
        and args.options.existence_check
        and not args.options.offline
        and not spec_file.exists()
        and (package_info := get_package_info(project.rpm_name))
    ):
        if args.options.compat:
            version_suffix = project.rpm_name.removeprefix(project.name)
            log.warn(
                f"Version {version_suffix}.* of the crate {project.name!r} is already "
                f"packaged for Fedora: {package_info['full_url']}",
            )
        else:
            log.warn(f"Crate {project.name!r} is already packaged for Fedora: {package_info['full_url']}")

        log.info("Re-run with --no-existence-check to create a new spec file from scratch.")
        sys.exit(1)


def license_file_check(project: Project, options: Options):
    if not project.license_files and not options.ignore_missing_license_files:
        log.error(
            "No license files were detected. "
            "In almost all cases, this is an issue with the upstream project that should be reported.",
        )
        log.info(
            "To temporarily ignore this issue, use the '--ignore-missing-license-files' / '-I' flag - "
            "for example, when manually including license files from the upstream project, or for test builds.",
        )
        sys.exit(1)


def handle_crate(project: CrateProject, args: CliArgs, conf: TomlConf) -> tuple[Path, str]:
    package = project.metadata.packages[0]
    spec_file = (args.output_directory or Path()) / f"{project.rpm_name}.spec"
    packager = detect_packager()

    license_file_check(project, args.options)
    existence_check(project, args, spec_file)

    # basic validation for extra sources and patches
    validate_extra_sources(conf, project.vendor_tarball)
    validate_extra_patches(
        conf,
        auto_patch=project.auto_patch is not None,
        manual_patch=project.manual_patch is not None,
    )

    if build_meta := Version.parse(package.version).build:
        log.error(f"Crate version {package.version!r} contains build metadata: '+{build_meta}'")
        log.error(f"This is not supported by rust2rpm; remove the '+{build_meta}' suffix.")
        sys.exit(1)

    if project.vendor_tarball and not (package.is_bin() or package.is_cdylib()):
        log.error("Building library-only crates with vendored dependencies is not supported.")
        sys.exit(1)

    if conf.package.cargo_toml_patch_comments and project.manual_patch is None:
        log.error(
            "The rust2rpm.toml configuration file specifies comments for a Cargo.toml patch, "
            "but Cargo.toml was not patched.",
        )
        sys.exit(1)

    if conf.package.url:
        log.error(
            "The rust2rpm.toml configuration file specifies an override for the package URL, "
            "but this is not valid for crates packaged from crates.io.",
        )
        sys.exit(1)

    if conf.package.source_url:
        log.error(
            "The rust2rpm.toml configuration file specifies an override for the Source URL, "
            "but this is not valid for crates packaged from crates.io.",
        )
        sys.exit(1)

    warn_if_package_uses_restrictive_dependencies(package)

    try:
        generator = CrateTemplate(
            project=project,
            conf=conf,
            target=args.target,
            date=None,
            packager=packager,
            use_relative_license_paths=args.options.relative_license_paths,
            use_rpmautospec=use_rpmautospec(args, spec_file),
            make_changelog_entry=args.options.auto_changelog_entry,
        )
        spec_contents = generator.render()

    except InvalidMetadataError as exc:
        # this can happen for crates that do not specify a license
        log.error(str(exc))
        sys.exit(1)

    except ValueError as exc:
        # this can happen for weird version requirements in ancient crates
        # that cannot be translated correctly into RPM dependencies
        log.error(f"Could not generate spec file due to invalid version requirement: {exc!s}")
        sys.exit(1)

    return spec_file, spec_contents


def handle_epel8(project: CrateProject, args: CliArgs, conf: TomlConf) -> tuple[Path, str]:
    package = project.metadata.packages[0]
    spec_file = (args.output_directory or Path()) / f"{project.rpm_name}.spec"
    packager = detect_packager()

    license_file_check(project, args.options)
    existence_check(project, args, spec_file)

    # basic validation for extra sources and patches
    validate_extra_sources(conf, project.vendor_tarball)
    validate_extra_patches(
        conf,
        auto_patch=project.auto_patch is not None,
        manual_patch=project.manual_patch is not None,
    )

    if build_meta := Version.parse(package.version).build:
        log.error(f"Crate version {package.version!r} contains build metadata: '+{build_meta}'")
        log.error(f"This is not supported by rust2rpm; remove the '+{build_meta}' suffix.")
        sys.exit(1)

    if project.vendor_tarball and not (package.is_bin() or package.is_cdylib()):
        log.error("Building library-only crates with vendored dependencies is not supported.")
        sys.exit(1)

    if conf.package.cargo_toml_patch_comments and project.manual_patch is None:
        log.error(
            "The rust2rpm.toml configuration file specifies comments for a Cargo.toml patch, "
            "but Cargo.toml was not patched.",
        )
        sys.exit(1)

    if conf.package.url:
        log.error(
            "The rust2rpm.toml configuration file specifies an override for the package URL, "
            "but this is not valid for crates packaged from crates.io.",
        )
        sys.exit(1)

    if conf.package.source_url:
        log.error(
            "The rust2rpm.toml configuration file specifies an override for the Source URL, "
            "but this is not valid for crates packaged from crates.io.",
        )
        sys.exit(1)

    warn_if_package_uses_restrictive_dependencies(package)

    try:
        generator = EpelEightTemplate(
            project=project,
            conf=conf,
            date=None,
            packager=packager,
            use_rpmautospec=use_rpmautospec(args, spec_file),
            make_changelog_entry=args.options.auto_changelog_entry,
        )
        spec_contents = generator.render()

    except InvalidMetadataError as exc:
        # this can happen for crates that do not specify a license
        log.error(str(exc))
        sys.exit(1)

    except ValueError as exc:
        # this can happen for weird version requirements in ancient crates
        # that cannot be translated correctly into RPM dependencies
        log.error(f"Could not generate spec file due to invalid version requirement: {exc!s}")
        sys.exit(1)

    return spec_file, spec_contents


def handle_local(project: LocalProject, args: CliArgs, conf: TomlConf) -> tuple[Path, str]:
    package = project.metadata.packages[0]
    spec_file = (args.output_directory or Path()) / f"{project.rpm_name}.spec"
    packager = detect_packager()

    license_file_check(project, args.options)
    existence_check(project, args, spec_file)

    # basic validation for extra sources and patches
    validate_extra_sources(conf, project.vendor_tarball)
    validate_extra_patches(
        conf,
        auto_patch=project.auto_patch is not None,
        manual_patch=project.manual_patch is not None,
    )

    if build_meta := Version.parse(package.version).build:
        log.error(f"Crate version {package.version!r} contains build metadata: '+{build_meta}'")
        log.error(f"This is not supported by rust2rpm; remove the '+{build_meta}' suffix.")
        sys.exit(1)

    if conf.package.cargo_toml_patch_comments and project.manual_patch is None:
        log.error(
            "The rust2rpm.toml configuration file specifies comments for a Cargo.toml patch, "
            "but Cargo.toml was not patched.",
        )
        sys.exit(1)

    warn_if_package_uses_restrictive_dependencies(package)

    try:
        generator = ProjectTemplate(
            project=project,
            conf=conf,
            target=args.target,
            date=None,
            packager=packager,
            use_rpmautospec=use_rpmautospec(args, spec_file),
            make_changelog_entry=args.options.auto_changelog_entry,
        )
        spec_contents = generator.render()

    except InvalidMetadataError as exc:
        # this can happen for crates that do not specify a license
        log.error(str(exc))
        sys.exit(1)

    return spec_file, spec_contents


def handle_workspace(project: LocalProject, args: CliArgs, conf: TomlConf) -> tuple[Path, str]:
    spec_file = (args.output_directory or Path()) / f"{project.rpm_name}.spec"
    packager = detect_packager()

    license_file_check(project, args.options)
    existence_check(project, args, spec_file)

    # basic validation for extra sources and patches
    validate_extra_sources(conf, project.vendor_tarball)

    try:
        generator = WorkspaceTemplate(
            project=project,
            conf=conf,
            target=args.target,
            date=None,
            packager=packager,
            use_rpmautospec=use_rpmautospec(args, spec_file),
            make_changelog_entry=args.options.auto_changelog_entry,
        )
        spec_contents = generator.render()

    except InvalidMetadataError as exc:
        # this can happen for crates that do not specify a license
        log.error(str(exc))
        sys.exit(1)

    except WorkspaceError as exc:
        # this can happen for confusing workspace metadata or invalid hints
        log.error(str(exc))
        sys.exit(1)

    return spec_file, spec_contents


def print_deprecation_warnings(args: CliArgs):
    if args.options.relative_license_paths is True:
        log.warn(
            "The '--relative-license-paths' flag produces slightly unexpected results."
            "It is deprecated and will be removed in a future version of rust2rpm.",
        )

    if args.options.auto_changelog_entry is False:
        log.warn("The '--no-auto-changelog-entry' is deprecated and will be removed in a future version of rust2rpm.")


def print_migration_errors(args: CliArgs):
    if (pkgid := args.pkgid) and ("/" in pkgid):
        log.error("Paths are no longer accepted as positional arguments. Use the '--path' argument instead.")
        sys.exit(1)


def main(argv: list[str] | None = None):
    args = CliArgs.parse(argv)
    print_deprecation_warnings(args)
    print_migration_errors(args)

    # validate config and exit early if config is invalid
    conf = load_config(args.target, path=args.config_file, features=None, migrate=False)
    if args.options.validate_only:
        sys.exit(0)

    # force generation of a Cargo.toml patch if comments are configured
    args.options.patch = args.options.patch or (
        # only force opening an editor if comments are present but neither --patch nor --reuse-patch were specified
        len(conf.package.cargo_toml_patch_comments) > 0
        and not args.options.patch
        and not args.options.reuse_patch
    )

    project = load_project(args, conf)

    # validate config with known feature names
    if isinstance(project, CrateProject):
        feature_names = project.package.get_feature_names()
        conf = load_config(args.target, path=args.config_file, features=feature_names, migrate=True)
    else:
        conf = load_config(args.target, path=args.config_file, features=None, migrate=True)

    # special handling for epel8 target
    if args.target == Target.EPEL8 and isinstance(project, CrateProject):
        spec_file, spec_contents = handle_epel8(project, args, conf)

    # package for a workspace project
    elif isinstance(project, LocalProject) and project.metadata.is_workspace():
        spec_file, spec_contents = handle_workspace(project, args, conf)

    # package for non-crate, non-workspace project
    elif isinstance(project, LocalProject):
        spec_file, spec_contents = handle_local(project, args, conf)

    # package for an actual Rust crate
    elif isinstance(project, CrateProject):
        spec_file, spec_contents = handle_crate(project, args, conf)

    else:  # pragma nocover
        log.error("Invalid project - this should not happen.")
        sys.exit(1)

    # ensure output directory exists
    if out_dir := args.output_directory:
        out_dir.mkdir(parents=True, exist_ok=True)
    else:
        out_dir = Path()

    # write spec file
    with spec_file.open("w") as file:
        file.write(spec_contents)
    log.success(f"Generated: {file.name}")

    # write patch files
    for patch in [project.auto_patch, project.manual_patch]:
        if patch is not None:
            with (out_dir / patch.name).open("w") as file:
                file.write("\n".join(patch.contents) + "\n")
            log.success(f"Generated: {file.name}")


if __name__ == "__main__":  # pragma nocover
    try:
        main()
    except KeyboardInterrupt:
        sys.exit(0)
