"""Module containing functionality for the TOML-based configuration format."""

import json
import sys
import textwrap
import tomllib
from collections import defaultdict
from dataclasses import dataclass, field
from importlib import resources
from pathlib import Path

import jsonschema

from rust2rpm import log

from .common import ConfError

TOML_SCHEMA = json.loads(
    resources.files("rust2rpm").joinpath("rust2rpm.schema.json").read_text(),
)


def conf_comments_to_spec_comments(comments: list[str] | None) -> list[str]:
    """Format list of strings as pretty spec file comments.

    Any string in the list of comments that contains a newline character
    is treated as pre-formatted and no line-wrapping is applied to it.

    Arguments:
        comments: Comments as list of strings.

    Returns:
        Pre-formatted comments as list of lines.

    """
    if not comments:
        return []

    lines = []
    for comment in comments:
        # single-line comments: wrap to 80 columns
        if "\n" not in comment.strip():
            lines.extend(
                textwrap.wrap(
                    comment,
                    width=80,
                    initial_indent="# * ",
                    subsequent_indent="#   ",
                    break_long_words=False,
                    break_on_hyphens=False,
                ),
            )

        # multi-line comments: assume pre-formatted
        else:
            first, *rest = (line.strip() for line in comment.strip().splitlines())
            lines.append(f"# * {first}")
            lines.extend([f"#   {line}" for line in rest])

    return lines


@dataclass(frozen=True)
class Source:
    """Properties of extra Source files."""

    file: str
    number: int
    comments: list[str] = field(default_factory=list)

    @staticmethod
    def from_data(data: dict) -> "Source":
        """Initialize from TOML data directly."""
        return Source(**data)

    @property
    def comment_lines(self) -> list[str]:
        """Format comments as spec comments."""
        return conf_comments_to_spec_comments(self.comments)

    @property
    def whitespace(self) -> str:
        """Whitespace between "SourceX:" tag and the "file name"."""
        return " " * (16 - (len("Source") + len(str(self.number)) + 1))


@dataclass(frozen=True)
class Patch:
    """Properties of extra Patch files."""

    file: str
    number: int
    comments: list[str] = field(default_factory=list)

    @staticmethod
    def from_data(data: dict) -> "Patch":
        """Initialize from TOML data directly."""
        return Patch(**data)

    @property
    def comment_lines(self) -> list[str]:
        """Format comments as spec comments."""
        return conf_comments_to_spec_comments(self.comments)

    @property
    def whitespace(self) -> str:
        """Whitespace between "PatchX:" tag and the "file name"."""
        return " " * (16 - (len("Patch") + len(str(self.number)) + 1))


@dataclass(frozen=True)
class FileInEx:
    """File inclusion and exclusion rules."""

    include: list[str] | None = None
    exclude: list[str] | None = None


@dataclass(frozen=True)
class Package:
    """Collection of package-specific settings."""

    summary: str | None = None
    """Override for the generated RPM Summary tag."""

    description: str | None = None
    """Override for the generated RPM %description."""

    url: str | None = None
    """Override for the generated RPM URL tag."""

    source_url: str | None = None
    """Override for the RPM Source tag."""

    supported_arches: list[str] | None = None
    """List of architectures supported by the package.

    For crates, this is used to conditionally run `%build` and `%check` scriptlets.
    For other projects, it is used as the value of an `ExclusiveArch` tag.
    """

    suppress_cdylib_install_fixme: bool | None = None
    """Flag that controls injection of the "FIXME" comment for packages that contain "cdylib" targets."""

    bin_package_name: str | None = None
    """Override for the generated binary subpackage name."""

    cargo_install_bin: bool | None = None
    """Flag that controls injection of the `%cargo_install_bin` macro and its value."""

    cargo_install_lib: bool | None = None
    """Flag that controls injection of the `%cargo_install_lib` macro and its value."""

    debuginfo_level: int | None = None
    """Flag that controls injection of the `%rustflags_debuginfo` macro and its value."""

    cargo_toml_patch_comments: list[str] = field(default_factory=list)
    """List of comments associated with manually applied changes to Cargo.toml."""

    license_files: list[str] | FileInEx = field(default_factory=FileInEx)
    """Settings for overriding the results of crawling the project sources for license files."""

    doc_files: list[str] | FileInEx = field(default_factory=FileInEx)
    """Settings for overriding the results of crawling the project sources for documentation files."""

    extra_sources: list[Source] = field(default_factory=list)
    """Settings for including additional `Source` files."""

    extra_patches: list[Patch] = field(default_factory=list)
    """Settings for including additional `Patch` files."""

    extra_files: list[str] = field(default_factory=list)
    """Setting for including additional files in the `%files` list of the built binary subpackage."""

    exclude_crate_files: list[str] = field(default_factory=list)
    """Setting for excluding files from the `%files` list of the subpackage that contains the crate sources."""

    bin_renames: dict[str, str] = field(default_factory=dict)
    """Setting for renaming executables, usually to avoid file conflicts with other packages."""

    @staticmethod
    def from_data(data: dict) -> "Package":
        """Initialize from TOML data directly."""
        args = {key.replace("-", "_"): value for key, value in data.items()}

        if extra_sources := data.get("extra-sources"):
            args["extra_sources"] = [Source.from_data(source) for source in extra_sources]

        if extra_patches := data.get("extra-patches"):
            args["extra_patches"] = [Patch.from_data(patch) for patch in extra_patches]

        if lf_obj := data.get("license-files"):
            args["license_files"] = FileInEx(**lf_obj) if isinstance(lf_obj, dict) else lf_obj

        if df_obj := data.get("doc-files"):
            args["doc_files"] = FileInEx(**df_obj) if isinstance(df_obj, dict) else df_obj

        return Package(**args)

    @property
    def cargo_toml_patch_comment_lines(self) -> list[str]:
        """Format comments as spec comments."""
        return conf_comments_to_spec_comments(self.cargo_toml_patch_comments)


@dataclass(frozen=True)
class PrePostScripts:
    """Collection of extra commands for an RPM package scriptlet."""

    pre: list[str] = field(default_factory=list)
    """List of additional commands that are injected before the `%cargo_*` macro in this scriptlet."""

    post: list[str] = field(default_factory=list)
    """List of additional commands that are injected after the `%cargo_*` macro in this scriptlet."""


@dataclass(frozen=True)
class Scripts:
    """Collection of extra commands for RPM package scriptlets."""

    prep: PrePostScripts = field(default_factory=PrePostScripts)
    """Additional commands for the `%prep` scriptlet."""

    build: PrePostScripts = field(default_factory=PrePostScripts)
    """Additional commands for the `%build` scriptlet."""

    install: PrePostScripts = field(default_factory=PrePostScripts)
    """Additional commands for the `%install` scriptlet."""

    check: PrePostScripts = field(default_factory=PrePostScripts)
    """Additional commands for the `%check` scriptlet."""

    @staticmethod
    def from_data(data: dict) -> "Scripts":
        """Initialize from TOML data directly."""
        prep = PrePostScripts(**value) if (value := data.get("prep")) else PrePostScripts()
        build = PrePostScripts(**value) if (value := data.get("build")) else PrePostScripts()
        install = PrePostScripts(**value) if (value := data.get("install")) else PrePostScripts()
        check = PrePostScripts(**value) if (value := data.get("check")) else PrePostScripts()

        return Scripts(prep, build, install, check)


@dataclass(frozen=True)
class TestsSkip:
    """Collections of names that are included as `--skip` arguments in `%cargo_test` macro arguments."""

    lib: list[str] = field(default_factory=list)
    bin: list[str] = field(default_factory=list)
    doc: list[str] = field(default_factory=list)
    bins: list[str] = field(default_factory=list)
    tests: list[str] = field(default_factory=list)


@dataclass(frozen=True)
class TestsSkipExact:
    """Collection of flags that control the inclusion of the `--exact` flag in `%cargo_test` macro arguments."""

    lib: bool = False
    bin: bool = False
    doc: bool = False
    bins: bool = False
    tests: bool = False


@dataclass(frozen=True)
class TestsComments:
    """Collections of strings that are included as pre-formatted comments for `%cargo_test` macro invocations."""

    lib: list[str] = field(default_factory=list)
    bin: list[str] = field(default_factory=list)
    doc: list[str] = field(default_factory=list)
    bins: list[str] = field(default_factory=list)
    tests: list[str] = field(default_factory=list)


@dataclass(frozen=True)
class Tests:
    """Collection of test-specific settings."""

    run: str | list[str] = field(default_factory=list)
    """Setting that controls which kinds of cargo tests are run."""

    skip: list[str] | TestsSkip = field(default_factory=list)
    """Setting that controls which tests are skipped."""

    skip_exact: bool | TestsSkipExact = False
    """Setting that controls whether exact matches or substring matches are used for determining skipped tests."""

    comments: list[str] | TestsComments = field(default_factory=list)
    """Collections of strings that are included as pre-formatted comments."""

    @staticmethod
    def from_data(data: dict) -> "Tests":
        """Initialize from TOML data directly."""
        args = {key.replace("-", "_"): value for key, value in data.items()}

        # validate the "tests.run" setting:
        # both "all" and "none" cannot be combimed with other values
        if (tests_run := data.get("run")) and isinstance(tests_run, list):
            if ("all" in tests_run or "none" in tests_run) and len(tests_run) != 1:
                msg = f"Invalid set of tests to run: {tests_run!r}"
                raise ConfError(msg)
        else:
            tests_run = []

        if (tests_skip := data.get("skip")) and isinstance(tests_skip, dict):
            args["skip"] = TestsSkip(**tests_skip)

            # validate the "tests.skip" setting:
            # skipping tests for targets that are not separately run has no effect
            for key in tests_skip:
                if key not in tests_run:
                    log.warn(
                        f"Skipping tests for {key!r} targets only has no effect if they are not run separately.",
                    )

        if (tests_skip_exact := data.get("skip-exact")) and isinstance(tests_skip_exact, dict):
            args["skip_exact"] = TestsSkipExact(**tests_skip_exact)

            # validate the "tests.skip-exact" setting:
            # exact test name matching for targets that are not separately run has no effect
            for key in tests_skip_exact:
                if key not in tests_run:
                    log.warn(
                        f"Using exact test name matches for {key!r} targets only has no effect "
                        "if they are not run separately.",
                    )

        if (tests_comments := data.get("comments")) and isinstance(tests_comments, dict):
            args["comments"] = TestsComments(**tests_comments)

            # validate the "tests.comments" setting:
            # specifying comments for test targets that are not separately run has no effect
            for key in tests_comments:
                if key not in tests_run:
                    log.warn(
                        f"Specifying comments for {key!r} target tests only has no effect "
                        "if they are not run separately.",
                    )

        return Tests(**args)


@dataclass(frozen=True)
class Features:
    """Collection of feature-specific settings."""

    enable_all: bool = False
    """Flag that controls whether the `--all-features` flag is passed to cargo."""

    enable: list[str] = field(default_factory=list)
    """List of features that are enabled and passed with `--features` to cargo."""

    disable_default: bool = False
    """Flag that controls whether the `--no-default-features` flag is passed to cargo."""

    hide: list[str] = field(default_factory=list)
    """List of features for which the generation of the corresponding subpackage is suppressed."""

    @staticmethod
    def from_data(data: dict, validate_features: set[str] | None = None) -> "Features":
        """Initialize from TOML data directly."""
        # validate the "features.enable-all" and "features.disable-default" settings:
        # both cannot be enabled at the same time
        if data.get("enable-all") is True and data.get("disable-default") is True:
            msg = "Conflicting settings for features: 'enable-all' and 'disable-default'"
            raise ConfError(msg)

        if validate_features is not None:
            # validate the "features.enable" setting:
            # list elements must be valid feature names
            if enable_list := data.get("enable"):
                for enabled in enable_list:
                    if enabled not in validate_features:
                        msg = f"Unrecognized enabled feature: {enabled}"
                        raise ConfError(msg)

            # validate the "features.hide" setting:
            # list elements must be valid feature names
            if hide_list := data.get("hide"):
                for hidden in hide_list:
                    if hidden not in validate_features:
                        msg = f"Unrecognized hidden feature: {hidden}"
                        raise ConfError(msg)
                if "default" in hide_list:
                    msg = "The 'default' feature is implicitly defined by cargo, it cannot be hidden."
                    raise ConfError(msg)

        # warn when conflicting settings are used
        if (data.get("enable-all") is True) and (data.get("hide") is not None and len(data["hide"]) > 0):
            log.warn(
                "Conflicting settings for features: "
                "All features are enabled for the build but some feature subpackages are hidden. "
                "This is likely an error.",
            )

        args = {key.replace("-", "_"): value for key, value in data.items()}
        return Features(**args)


@dataclass(frozen=True)
class Requires:
    """Collection of settings related to additional Requires and BuildRequires."""

    build: list[str] = field(default_factory=list)
    """List of additional `BuildRequires`."""

    test: list[str] = field(default_factory=list)
    """List of additional `BuildRequires` that only apply when building the package with tests."""

    lib: list[str] = field(default_factory=list)
    """List of additional `Requires` for the subpackage that contains the crate sources."""

    bin: list[str] = field(default_factory=list)
    """List of additional `Requires` for the built binary subpackage."""

    features: defaultdict[str, list[str]] = field(default_factory=lambda: defaultdict(list))
    """List of additional `Requires` for built subpackages associated with crate features."""

    @staticmethod
    def from_data(data: dict, validate_features: set[str] | None = None) -> "Requires":
        """Initialize from TOML data directly."""
        args = {key.replace("-", "_"): value for key, value in data.items()}

        # validate the "requires.features" setting
        # dictionary keys must be valid feature names
        if validate_features is not None and (feature_requires := data.get("features")):
            for feature in feature_requires:
                if feature not in validate_features:
                    msg = f"Unrecognized Requires for feature: {feature}"
                    raise ConfError(msg)

        return Requires(**args)


@dataclass(frozen=True)
class TomlConf:
    """rust2rpm configuration loaded from the rust2rpm.toml file format."""

    package: Package = field(default_factory=Package)
    """Settings from the [package] table."""

    scripts: Scripts = field(default_factory=Scripts)
    """Settings from the [scripts] table."""

    tests: Tests = field(default_factory=Tests)
    """Settings from the [tests] table."""

    features: Features = field(default_factory=Features)
    """Settings from the [features] table."""

    requires: Requires = field(default_factory=Requires)
    """Settings from the [requires] table."""

    @staticmethod
    def from_data(data: dict, validate_features: set[str] | None) -> "TomlConf":
        """Initialize from TOML data directly."""
        jsonschema.validate(data, TOML_SCHEMA)

        # the "default" feature is always implicitly defined
        if validate_features is not None and "default" not in validate_features:
            validate_features.add("default")

        package = Package.from_data(data.get("package") or {})
        scripts = Scripts.from_data(data.get("scripts") or {})
        tests = Tests.from_data(data.get("tests") or {})

        features = Features.from_data(data.get("features") or {}, validate_features)
        requires = Requires.from_data(data.get("requires") or {}, validate_features)

        return TomlConf(package, scripts, tests, features, requires)

    @staticmethod
    def from_str(value: str, validate_features: set[str] | None) -> "TomlConf":
        """Load configuration from a TOML string."""
        toml = tomllib.loads(value)
        return TomlConf.from_data(toml, validate_features)

    @staticmethod
    def load(path: Path, validate_features: set[str] | None) -> "TomlConf":
        """Load configuration from a file path."""
        with path.open() as file:
            contents = file.read()

        return TomlConf.from_str(contents, validate_features)


def load_config_toml(path: Path, validate_features: set[str] | None = None) -> TomlConf | None:
    """Load rust2rpm configuration from the rust2rpm.toml file format.

    Arguments:
        path: Expected file path of the rust2rpm.toml file.
        validate_features: Optional set of valid feature names.
            Used to validate some settings.

    Returns:
        Configuration loaded from a file in TOML format, or `None` if no file was found.

    Raises:
        `SystemExit` on fatal errors.

    """
    try:
        tomlconf = TomlConf.load(path, validate_features)

    except FileNotFoundError:
        return None

    except tomllib.TOMLDecodeError as exc:
        log.error("Cannot read rust2rpm.toml file (TOML syntax error):")
        log.error(str(exc))
        sys.exit(1)

    except jsonschema.ValidationError as exc:
        if not exc.path:
            log.error("Invalid rust2rpm.toml file (unknown setting or table):")
            log.error(exc.message)
        else:
            err_path = ""
            for elem in exc.path:
                if isinstance(elem, int):
                    err_path += f"[{elem}]"
                else:
                    err_path += f".{elem}"
            log.error(f"Invalid rust2rpm.toml file (invalid setting at {err_path!r}):")
            log.error(exc.message)
        sys.exit(1)

    except ConfError as exc:
        log.error("Invalid rust2rpm.toml file:")
        log.error(str(exc))
        sys.exit(1)

    else:
        return tomlconf
