"""Module containing functionality for inspecting contents of crate files / tarballs."""

import contextlib
import os
import re
import tarfile
import tempfile
from collections.abc import Generator
from pathlib import Path
from typing import Any

from rust2rpm import log
from rust2rpm.conf import FileInEx, TomlConf

LICENSE_FILE_PATTERN = re.compile(
    r"""
    COPYING(?:[.-].*)?|COPYRIGHT(?:[.-].*)?|
    EULA(?:[.-].*)?|[Ll]icen[cs]e(?![.]rs$).*|
    (?:.*[.-])?(?:UN)?LICEN[CS]E(?:[.-].*)?|NOTICE(?:[.-].*)?|
    PATENTS(?:[.-].*)?|
    (?:agpl|l?gpl)[.-].*|CC-BY-.*|
    (?:AGPL|APACHE|BSD|GFDL|GNU|L?GPL|MIT|MPL|OFL)-.*[0-9].*
    """,
    re.VERBOSE,
)
"""Regular expression for heuristically determining files that contain license
texts based on the file name."""

LICENSE_EXCLUDE_DIRS = {
    "target",
    "vendor",
    "example",
    "examples",
    "_example",
    "_examples",
    "testdata",
    "_testdata",
    ".git",
    ".github",
    "tests",
    "test",
}
"""List of directories that are ignored when crawling project sources for
license files."""

DOC_FILE_PATTERN = re.compile(
    r"""
    .*\.(?:md|markdown|mdown|mkdn|rst|txt)|AUTHORS|
    AUTHORS[.-].*|CONTRIBUTORS|CONTRIBUTORS[.-].*|README|
    README[.-].*|CHANGELOG|CHANGELOG[.-].*|TODO|TODO[.-].*
    """,
    re.IGNORECASE | re.VERBOSE,
)
"""Regular expression for heuristically determining documentation files based
on the file name."""

DOC_FILE_EXCLUDES = re.compile(r"CMakeLists\.txt|versions\.txt|.*\.tpl|.*\.in")
"""Regular expression used to ignore files that look like documentation but
actually are not when crawling project sources."""


class UnsafeTarballError(Exception):
    """Raised if the downloaded crate file is a tarball with unsafe contents."""


@contextlib.contextmanager
def files_from_crate(
    crate_path: Path,
    crate_name: str,
    crate_version: str,
    tomlconf: TomlConf,
) -> Generator[tuple[Path, list[str], list[str]], Any, None]:
    """Unpack the crate at the given path and inspect its contents.

    Arguments:
        crate_path: Path to the crate file / tarball.
        crate_name: Name of the crate.
        crate_version: Version of the crate.
        tomlconf: Global rust2rpm configuration, which contains settings for
            filtering / overriding heuristics for license and documentation
            file detection.

    Yields:
        Three-tuple of (path to the `Cargo.toml` file inside unpacked sources,
        list of detected license files, and list of detected documentation files).

    """
    with tempfile.TemporaryDirectory() as tmpdir:
        target_dir = f"{tmpdir}/"

        with tarfile.open(crate_path, "r") as archive:
            for n in archive.getnames():

                if not str((Path(target_dir) / n).resolve()).startswith(target_dir):
                    msg = "Crate tar archive contains unsafe filenames"
                    raise UnsafeTarballError(msg)

            # S202: the loop above checks that all filenames are safe
            archive.extractall(target_dir)  # noqa: S202

        toml_path = Path(f"{tmpdir}/{crate_name}-{crate_version}/Cargo.toml")
        if not toml_path.is_file():
            msg = "Crate does not contain a Cargo.toml file."
            raise FileNotFoundError(msg)

        root_path = Path(f"{tmpdir}/{crate_name}-{crate_version}")
        license_files = get_license_files(root_path, tomlconf)
        doc_files = get_doc_files(root_path, tomlconf)

        yield toml_path, license_files, doc_files


def filter_files_list(kind: str, path: Path, overrides: list[str]) -> list[str]:
    """Override detected files with the given list of files and / or globs.

    Arguments:
        kind: The kind of file currently being processed (`"license"` or `"doc"`).
        path: Root directory of the project sources.
        overrides: List of file names or globs. Plain file names are used as-is,
            globs are expanded inside the project root directory. If no files
            match a specific glob, a warning is logged.

    Returns:
        Sorted list of file names relative to the project root directory.

    """
    results: set[str] = set()

    for entry in overrides:
        if "*" in entry or "?" in entry:
            matches = sorted(path.glob(entry))
            if not matches:
                log.warn(f"No matches for specified %{kind} file glob: {entry!r}")
            results.update(str(match.relative_to(path)) for match in matches)
        else:
            results.add(entry)

    return sorted(results)


def filter_files_in_ex(kind: str, path: Path, files: set[str], overrides: FileInEx) -> list[str]:
    """Filter detected files given specific include and exclude settings.

    **Includes** are processed first - plain file names are added to the list
    of files as-is, and a warning is logged if an explicitly included file was
    already in the set of heuristically detected files. Globs are expanded
    inside the project root directory and any matches are added to the list.
    If no files match a specific glob, a warning is logged.

    **Excludes** are processed second - plain file names are removed from the
    list of files as-is, and a warning is logged if an explicitly excluded file
    was already not in the set of heuristically detected files. Globs are
    expanded inside the project root directory and any matches are removed from
    the list. If no files match a specific glob, a warning is logged.

    It is possible for exclude settings to remove files that were explicitly
    added with include settings when using conflicting settings.

    Arguments:
        kind: The kind of file currently being processed (`"license"` or `"doc"`).
        path: Root directory of the project sources.
        files: Heuristically determined set of files.
        overrides: List of file names or globs for both include and exclude
            settings.

    Returns:
        Sorted list of file names relative to the project root directory.

    """
    results = files.copy()

    include_list: list[str] = []
    exclude_list: list[str] = []

    for include in overrides.include or []:
        if "*" in include or "?" in include:
            matches = sorted(path.glob(include))
            if not matches:
                log.warn(f"No matches for specified %{kind} file glob: {include!r}")
            include_list.extend(str(match.relative_to(path)) for match in matches)
        else:
            if include in results:
                log.warn(f"Manually included %{kind} file already detected: {include!r}")
                continue
            include_list.append(include)

    for exclude in overrides.exclude or []:
        if "*" in exclude or "?" in exclude:
            matches = sorted(path.glob(exclude))
            if not matches:
                log.warn(f"No matches for specified %{kind} file glob: {exclude!r}")
            exclude_list.extend(str(match.relative_to(path)) for match in matches)
        else:
            if exclude not in results:
                log.warn(f"Manually excluded %{kind} file not detected: {exclude!r}")
                continue
            exclude_list.append(exclude)

    results.update(include_list)
    results.difference_update(exclude_list)

    return sorted(results)


def filter_files(kind: str, path: Path, files: set[str], overrides: list[str] | FileInEx) -> list[str]:
    """Filter detected files given specific overrides or include and exclude settings.

    This function just calls `filter_files_list` or `filter_files_in_ex`
    depending on the type of the `overrides` argument.

    Arguments:
        kind: The kind of file currently being processed (`"license"` or `"doc"`).
        path: Root directory of the project sources.
        files: Heuristically determined set of files.
        overrides: List of file names or globs for both include and exclude
            settings, or include / exclude settings.

    Returns:
        Sorted list of file names relative to the project root directory.

    """
    if isinstance(overrides, list):
        return filter_files_list(kind, path, overrides)

    if isinstance(overrides, FileInEx):
        return filter_files_in_ex(kind, path, files, overrides)

    raise ValueError  # pragma nocover


def get_license_files(path: Path, tomlconf: TomlConf) -> list[str]:
    """Crawl the given path for license files and apply overrides / filtering rules.

    Arguments:
        path: Root directory of the project sources.
        tomlconf: Global rust2rpm configuration, which contains settings for
            filtering / overriding heuristics for license and documentation
            file detection.

    Returns:
        Sorted list of file names relative to the project root directory.

    """
    results: set[str] = set()

    for root, dirs, files in os.walk(path, topdown=True):
        dirs[:] = [d for d in dirs if d not in LICENSE_EXCLUDE_DIRS]
        for f in files:
            if LICENSE_FILE_PATTERN.match(f):
                results.add(str((Path(root) / f).relative_to(path)))

    return filter_files("license", path, results, tomlconf.package.license_files)


def get_doc_files(path: Path, tomlconf: TomlConf) -> list[str]:
    """Crawl the given path for documentation files and apply overrides / filtering rules.

    Arguments:
        path: Root directory of the project sources.
        tomlconf: Global rust2rpm configuration, which contains settings for
            filtering / overriding heuristics for license and documentation
            file detection.

    Returns:
        Sorted list of file names relative to the project root directory.

    """
    results: set[str] = set()

    for root, dirs, files in os.walk(path, topdown=True):
        dirs[:] = []
        for f in files:
            if (
                DOC_FILE_PATTERN.fullmatch(f)
                and not LICENSE_FILE_PATTERN.fullmatch(f)
                and not DOC_FILE_EXCLUDES.fullmatch(f)
            ):
                relpath = str((Path(root) / f).relative_to(path))
                if not relpath.startswith("target/"):
                    results.add(relpath)

    return filter_files("doc", path, results, tomlconf.package.doc_files)
